from typing import Any, Dict, Type, Sequence

from golem.core.dag.graph import Graph
from golem.core.dag.graph_delegate import GraphDelegate
from golem.core.dag.linked_graph_node import LinkedGraphNode


def graph_from_json(cls: Type[Graph], json_obj: Dict[str, Any]) -> Graph:
    obj = cls()

    nodes_key = 'nodes' if 'nodes' in json_obj else '_nodes'
    if not issubclass(cls, GraphDelegate):
        nodes = json_obj[nodes_key]
        _reassign_edges_by_node_ids(nodes)
        obj.nodes = nodes

    # GraphDelegate case is handled by this
    vars(obj).update(**{k: v for k, v in json_obj.items() if k != nodes_key})

    return obj


def _reassign_edges_by_node_ids(nodes: Sequence[LinkedGraphNode]):
    """
    Assigns each <inner_node> from "nodes_from" to equal <outer_node> from "nodes"
        (cause each node from "nodes_from" in fact should point to the same node from "nodes")
    """
    lookup_dict = {node.uid: node for node in nodes}
    for node in nodes:
        if node.nodes_from:
            for parent_node_idx, parent_node_uid in enumerate(node.nodes_from):
                node.nodes_from[parent_node_idx] = lookup_dict.get(parent_node_uid, None)



from functools import partial
from typing import (Callable, TYPE_CHECKING)

from golem.core.optimisers.genetic.operators.operator import PopulationT, Operator
from golem.core.optimisers.genetic.operators.selection import Selection
from golem.core.utilities.data_structures import ComparableEnum as Enum

if TYPE_CHECKING:
    from golem.core.optimisers.genetic.gp_params import GPAlgorithmParameters


class GeneticSchemeTypesEnum(Enum):
    steady_state = 'steady_state'
    generational = 'generational'
    parameter_free = 'parameter_free'


class Inheritance(Operator):
    def __init__(self, parameters: 'GPAlgorithmParameters',
                 selection: Selection):
        super().__init__(parameters=parameters)
        self.selection = selection

    def __call__(self, previous_population: PopulationT, new_population: PopulationT) -> PopulationT:
        return self._inheritance_type_by_genetic_scheme(previous_population, new_population)()

    def _inheritance_type_by_genetic_scheme(self, previous_population: PopulationT,
                                            new_population: PopulationT) -> Callable:
        steady_state_scheme = partial(steady_state_inheritance, previous_population,
                                      new_population, self.selection)
        generational_scheme = partial(direct_inheritance, new_population, self.parameters.pop_size)
        inheritance_type_by_genetic_scheme = {
            GeneticSchemeTypesEnum.generational: generational_scheme,
            GeneticSchemeTypesEnum.steady_state: steady_state_scheme,
            GeneticSchemeTypesEnum.parameter_free: steady_state_scheme
        }
        return inheritance_type_by_genetic_scheme[self.parameters.genetic_scheme_type]


def steady_state_inheritance(prev_population: PopulationT, new_population: PopulationT, selection: Selection) \
        -> PopulationT:
    selected_individuals = selection.individuals_selection(individuals=prev_population + new_population)
    return selected_individuals


def direct_inheritance(new_population: PopulationT, pop_size: int) -> PopulationT:
    return new_population[:pop_size]



from typing import Union, Optional, Sequence

from golem.core.dag.graph import Graph
from golem.core.optimisers.genetic.gp_params import GPAlgorithmParameters
from golem.core.optimisers.genetic.operators.mutation import Mutation
from golem.core.optimisers.objective import Objective
from golem.core.optimisers.opt_history_objects.individual import Individual
from golem.core.optimisers.optimization_parameters import GraphRequirements
from golem.core.optimisers.optimizer import GraphGenerationParams
from golem.core.optimisers.random.random_search import RandomSearchOptimizer


class RandomMutationSearchOptimizer(RandomSearchOptimizer):
    """
    Random search-based graph models optimizer
    """

    def __init__(self,
                 objective: Objective,
                 initial_graphs: Union[Graph, Sequence[Graph]],
                 requirements: Optional[GraphRequirements] = None,
                 graph_generation_params: Optional[GraphGenerationParams] = None,
                 graph_optimizer_params: Optional[GPAlgorithmParameters] = None):
        requirements = requirements or GraphRequirements()
        graph_optimizer_params = graph_optimizer_params or GPAlgorithmParameters()
        super().__init__(objective, initial_graphs, requirements, graph_generation_params, graph_optimizer_params)
        self.mutation = Mutation(self.graph_optimizer_params, self.requirements, self.graph_generation_params)

    def _generate_new_individual(self) -> Individual:
        return self.mutation(self.best_individual)



from __future__ import annotations

import datetime
import os
from copy import deepcopy
from pathlib import Path
from textwrap import wrap
from typing import TYPE_CHECKING, Any, Callable, Dict, Iterable, Optional, Sequence, Tuple, Union, List
from uuid import uuid4

import networkx as nx
import numpy as np
from matplotlib import pyplot as plt
from matplotlib.colors import to_hex
from matplotlib.patches import ArrowStyle
from pyvis.network import Network
from seaborn import color_palette

from golem.core.dag.graph_utils import distance_to_primary_level
from golem.core.dag.convert import graph_structure_as_nx_graph
from golem.core.log import default_log
from golem.core.paths import default_data_dir

if TYPE_CHECKING:
    from golem.core.dag.graph import Graph
    from golem.core.optimisers.graph import OptGraph

    GraphType = Union[Graph, OptGraph]

PathType = Union[os.PathLike, str]

MatplotlibColorType = Union[str, Sequence[float]]
LabelsColorMapType = Dict[str, MatplotlibColorType]
NodeColorFunctionType = Callable[[Iterable[str]], LabelsColorMapType]
NodeColorType = Union[MatplotlibColorType, LabelsColorMapType, NodeColorFunctionType]


class GraphVisualizer:
    def __init__(self, graph: GraphType, visuals_params: Optional[Dict[str, Any]] = None, ):
        visuals_params = visuals_params or {}
        default_visuals_params = dict(
            engine='matplotlib',
            dpi=100,
            node_color=self.__get_colors_by_labels,
            node_size_scale=1.0,
            font_size_scale=1.0,
            edge_curvature_scale=1.0,
            graph_to_nx_convert_func=graph_structure_as_nx_graph
        )
        default_visuals_params.update(visuals_params)
        self.visuals_params = default_visuals_params
        self.graph = graph

        self.log = default_log(self)

    def visualise(self, save_path: Optional[PathType] = None, engine: Optional[str] = None,
                  node_color: Optional[NodeColorType] = None, dpi: Optional[int] = None,
                  node_size_scale: Optional[float] = None,
                  font_size_scale: Optional[float] = None, edge_curvature_scale: Optional[float] = None,
                  nodes_labels: Dict[int, str] = None, edges_labels: Dict[int, str] = None):
        engine = engine or self.get_predefined_value('engine')

        if not self.graph.nodes:
            raise ValueError('Empty graph can not be visualized.')

        if engine == 'matplotlib':
            self.__draw_with_networkx(save_path=save_path, node_color=node_color, dpi=dpi,
                                      node_size_scale=node_size_scale, font_size_scale=font_size_scale,
                                      edge_curvature_scale=edge_curvature_scale,
                                      nodes_labels=nodes_labels, edges_labels=edges_labels)
        elif engine == 'pyvis':
            self.__draw_with_pyvis(save_path, node_color)
        elif engine == 'graphviz':
            self.__draw_with_graphviz(save_path, node_color, dpi)
        else:
            raise NotImplementedError(f'Unexpected visualization engine: {engine}. '
                                      'Possible values: matplotlib, pyvis, graphviz.')

    @staticmethod
    def __get_colors_by_labels(labels: Iterable[str]) -> LabelsColorMapType:
        unique_labels = list(set(labels))
        palette = color_palette('tab10', len(unique_labels))
        return {label: palette[unique_labels.index(label)] for label in labels}

    def __draw_with_graphviz(self, save_path: Optional[PathType] = None, node_color: Optional[NodeColorType] = None,
                             dpi: Optional[int] = None, graph_to_nx_convert_func: Optional[Callable] = None):
        save_path = save_path or self.get_predefined_value('save_path')
        node_color = node_color or self.get_predefined_value('node_color')
        dpi = dpi or self.get_predefined_value('dpi')
        graph_to_nx_convert_func = graph_to_nx_convert_func or self.get_predefined_value('graph_to_nx_convert_func')

        nx_graph, nodes = graph_to_nx_convert_func(self.graph)
        # Define colors
        if callable(node_color):
            colors = node_color([str(node) for node in nodes.values()])
        elif isinstance(node_color, dict):
            colors = node_color
        else:
            colors = {str(node): node_color for node in nodes.values()}
        for n, data in nx_graph.nodes(data=True):
            label = str(nodes[n])
            data['label'] = label.replace('_', ' ')
            data['color'] = to_hex(colors.get(label, colors.get(None)))

        gv_graph = nx.nx_agraph.to_agraph(nx_graph)
        kwargs = {'prog': 'dot', 'args': f'-Gnodesep=0.5 -Gdpi={dpi} -Grankdir="LR"'}

        if save_path:
            gv_graph.draw(save_path, **kwargs)
        else:
            save_path = Path(default_data_dir(), 'graph_plots', str(uuid4()) + '.png')
            save_path.parent.mkdir(exist_ok=True)
            gv_graph.draw(save_path, **kwargs)

            img = plt.imread(str(save_path))
            plt.imshow(img)
            plt.gca().axis('off')
            plt.gcf().set_dpi(dpi)
            plt.tight_layout()
            plt.show()
            remove_old_files_from_dir(save_path.parent)

    def __draw_with_pyvis(self, save_path: Optional[PathType] = None, node_color: Optional[NodeColorType] = None,
                          graph_to_nx_convert_func: Optional[Callable] = None):
        save_path = save_path or self.get_predefined_value('save_path')
        node_color = node_color or self.get_predefined_value('node_color')
        graph_to_nx_convert_func = graph_to_nx_convert_func or self.get_predefined_value('graph_to_nx_convert_func')

        net = Network('500px', '1000px', directed=True)
        nx_graph, nodes = graph_to_nx_convert_func(self.graph)
        # Define colors
        if callable(node_color):
            colors = node_color([str(node) for node in nodes.values()])
        elif isinstance(node_color, dict):
            colors = node_color
        else:
            colors = {str(node): node_color for node in nodes.values()}
        for n, data in nx_graph.nodes(data=True):
            operation = nodes[n]
            label = str(operation)
            data['n_id'] = str(n)
            data['label'] = label.replace('_', ' ')
            params = operation.content.get('params')
            if isinstance(params, dict):
                params = str(params)[1:-1]
            data['title'] = params
            data['level'] = distance_to_primary_level(operation)
            data['color'] = to_hex(colors.get(label, colors.get(None)))
            data['font'] = '20px'
            data['labelHighlightBold'] = True

        for _, data in nx_graph.nodes(data=True):
            net.add_node(**data)
        for u, v in nx_graph.edges:
            net.add_edge(str(u), str(v))

        if save_path:
            net.save_graph(str(save_path))
            return
        save_path = Path(default_data_dir(), 'graph_plots', str(uuid4()) + '.html')
        save_path.parent.mkdir(exist_ok=True)
        net.show(str(save_path))
        remove_old_files_from_dir(save_path.parent)

    def __draw_with_networkx(self, save_path: Optional[PathType] = None,
                             node_color: Optional[NodeColorType] = None,
                             dpi: Optional[int] = None, node_size_scale: Optional[float] = None,
                             font_size_scale: Optional[float] = None, edge_curvature_scale: Optional[float] = None,
                             graph_to_nx_convert_func: Optional[Callable] = None,
                             nodes_labels: Dict[int, str] = None, edges_labels: Dict[int, str] = None):
        save_path = save_path or self.get_predefined_value('save_path')
        node_color = node_color or self.get_predefined_value('node_color')
        dpi = dpi or self.get_predefined_value('dpi')
        node_size_scale = node_size_scale or self.get_predefined_value('node_size_scale')
        font_size_scale = font_size_scale or self.get_predefined_value('font_size_scale')
        edge_curvature_scale = (edge_curvature_scale if edge_curvature_scale is not None
                                else self.get_predefined_value('edge_curvature_scale'))
        graph_to_nx_convert_func = graph_to_nx_convert_func or self.get_predefined_value('graph_to_nx_convert_func')

        fig, ax = plt.subplots(figsize=(7, 7))
        fig.set_dpi(dpi)

        self.draw_nx_dag(self.graph, ax, node_color, node_size_scale, font_size_scale, edge_curvature_scale,
                         graph_to_nx_convert_func, nodes_labels, edges_labels)
        if not save_path:
            plt.show()
        else:
            plt.savefig(save_path, dpi=dpi)
            plt.close()

    def draw_nx_dag(self, graph: GraphType, ax: Optional[plt.Axes] = None,
                    node_color: Optional[NodeColorType] = None,
                    node_size_scale: float = 1, font_size_scale: float = 1, edge_curvature_scale: float = 1,
                    graph_to_nx_convert_func: Callable = graph_structure_as_nx_graph,
                    nodes_labels: Dict[int, str] = None, edges_labels: Dict[int, str] = None):

        def draw_nx_labels(pos, node_labels, ax, max_sequence_length, font_size_scale=1.0):
            def get_scaled_font_size(nodes_amount):
                min_size = 2
                max_size = 20

                size = min_size + int((max_size - min_size) / np.log2(max(nodes_amount, 2)))
                return size

            if ax is None:
                ax = plt.gca()
            for node, (x, y) in pos.items():
                text = '\n'.join(wrap(node_labels[node].replace('_', ' ').replace('-', ' '), 10))
                ax.text(x, y, text, ha='center', va='center',
                        fontsize=get_scaled_font_size(max_sequence_length) * font_size_scale,
                        bbox=dict(alpha=0.9, color='w', boxstyle='round'))

        def get_scaled_node_size(nodes_amount):
            min_size = 500
            max_size = 5000
            size = min_size + int((max_size - min_size) / np.log2(max(nodes_amount, 2)))
            return size

        if ax is None:
            ax = plt.gca()

        nx_graph, nodes = graph_to_nx_convert_func(graph)
        # Define colors
        if callable(node_color):
            node_color = node_color([str(node) for node in nodes.values()])
        if isinstance(node_color, dict):
            node_color = [node_color.get(str(node), node_color.get(None)) for node in nodes.values()]
        # Define hierarchy_level
        for node_id, node_data in nx_graph.nodes(data=True):
            node_data['hierarchy_level'] = distance_to_primary_level(nodes[node_id])
        # Get nodes positions
        pos, longest_sequence = get_hierarchy_pos(nx_graph)
        node_size = get_scaled_node_size(longest_sequence) * node_size_scale
        # Draw the graph's nodes.
        nx.draw_networkx_nodes(nx_graph, pos, node_size=node_size, ax=ax, node_color='w', linewidths=3,
                               edgecolors=node_color)
        # Draw the graph's node labels.
        draw_nx_labels(pos, {node_id: str(node) for node_id, node in nodes.items()}, ax, longest_sequence,
                       font_size_scale)
        # The ongoing section defines curvature for all edges.
        #   This is 'connection style' for an edge that does not intersect any nodes.
        connection_style = 'arc3'
        #   This is 'connection style' template for an edge that is too close to any node and must bend around it.
        #   The curvature value is defined individually for each edge.
        connection_style_curved_template = connection_style + ',rad={}'
        default_edge_curvature = 0.3
        #   The minimum distance from a node to an edge on which the edge must bend around the node.
        node_distance_gap = 0.15
        for u, v, e in nx_graph.edges(data=True):
            e['connectionstyle'] = connection_style
            p_1, p_2 = np.array(pos[u]), np.array(pos[v])
            p_1_2 = p_2 - p_1
            p_1_2_length = np.linalg.norm(p_1_2)
            # Finding the closest node to the edge.
            min_distance_found = node_distance_gap * 2  # It just must be bigger than the gap.
            closest_node_id = None
            for node_id in nx_graph.nodes:
                if node_id in (u, v):
                    continue  # The node is adjacent to the edge.
                p_3 = np.array(pos[node_id])
                distance_to_node = abs(np.cross(p_1_2, p_3 - p_1)) / p_1_2_length
                if (distance_to_node > min(node_distance_gap, min_distance_found) or  # The node is too far.
                        ((p_3 - p_1) @ p_1_2) < 0 or  # There's no perpendicular from the node to the edge.
                        ((p_3 - p_2) @ -p_1_2) < 0):
                    continue
                min_distance_found = distance_to_node
                closest_node_id = node_id

            if closest_node_id is None:
                continue  # There's no node to bend around.
            # Finally, define the edge's curvature based on the closest node position.
            p_3 = np.array(pos[closest_node_id])
            p_1_3 = p_3 - p_1
            curvature_strength = default_edge_curvature * edge_curvature_scale
            # 'alpha' denotes the angle between the abscissa and the edge.
            cos_alpha = p_1_2[0] / p_1_2_length
            sin_alpha = np.sqrt(1 - cos_alpha ** 2) * (-1) ** (p_1_2[1] < 0)
            # The closest node is placed as if the edge matched the abscissa.
            # Then, its ordinate shows on which side of the edge it is, "on the left" or "on the right".
            rotation_matrix = np.array([[cos_alpha, sin_alpha], [-sin_alpha, cos_alpha]])
            p_1_3_rotated = rotation_matrix @ p_1_3
            curvature_direction = (-1) ** (p_1_3_rotated[1] < 0)  # +1 is a "boat" \/, -1 is a "cat" /\.
            edge_curvature = curvature_direction * curvature_strength
            e['connectionstyle'] = connection_style_curved_template.format(edge_curvature)
        # Draw the graph's edges.
        arrow_style = ArrowStyle('Simple', head_length=1.5, head_width=0.8)
        for u, v, e in nx_graph.edges(data=True):
            nx.draw_networkx_edges(nx_graph, pos, edgelist=[(u, v)], node_size=node_size, ax=ax, arrowsize=10,
                                   arrowstyle=arrow_style, connectionstyle=e['connectionstyle'])
        if nodes_labels or edges_labels:
            self._set_labels(ax, pos, nx_graph,
                             longest_sequence, longest_sequence, font_size_scale,
                             nodes_labels, edges_labels)
        # Rescale the figure for all nodes to fit in.
        x_1, x_2 = ax.get_xlim()
        y_1, y_2 = ax.get_ylim()
        offset = 0.2
        x_offset = x_2 * offset
        y_offset = y_2 * offset
        ax.set_xlim(x_1 - x_offset, x_2 + x_offset)
        ax.set_ylim(y_1 - y_offset, y_2 + y_offset)
        ax.axis('off')
        plt.tight_layout()

    def get_predefined_value(self, param: str):
        return self.visuals_params.get(param)

    def _set_labels(self, ax: plt.Axes, pos: Any, nx_graph: nx.DiGraph,
                    longest_sequence: int, longest_y_sequence: int, font_size_scale: float,
                    nodes_labels: Dict[int, str], edges_labels: Dict[int, str]):
        """ Set labels with scores to nodes and edges. """

        def calculate_labels_bias(ax: plt.Axes, longest_y_sequence: int):
            y_1, y_2 = ax.get_ylim()
            y_size = y_2 - y_1
            if longest_y_sequence == 1:
                bias_scale = 0.25  # Fits between the central line and the upper bound.
            else:
                bias_scale = 1 / longest_y_sequence / 3 * 0.9  # Fits between the narrowest horizontal rows.
            bias = y_size * bias_scale
            return bias

        def _get_scaled_font_size(nodes_amount: int, size_scale: float) -> float:
            min_size = 11
            max_size = 25
            size = max(max_size * (1 - np.log10(nodes_amount)), min_size)
            return size * size_scale

        def match_labels_with_nx_nodes(nx_graph: nx.DiGraph, labels: Dict[int, str]) -> Dict[str, str]:
            """ Matches index of node in GOLEM graph with networkx node name. """
            nx_nodes = list(nx_graph.nodes.keys())
            nx_labels = {}
            for index in labels:
                nx_labels[nx_nodes[index]] = labels[index]
            return nx_labels

        def match_labels_with_nx_edges(nx_graph: nx.DiGraph, labels: Dict[int, str]) \
                -> Dict[Tuple[str, str], List[str]]:
            """ Matches index of edge in GOLEM graph with tuple of networkx nodes names. """
            nx_nodes = list(nx_graph.nodes.keys())
            edges = self.graph.get_edges()
            nx_labels = {}
            for index in labels:
                edge = edges[index]
                parent_node_nx = nx_nodes[self.graph.nodes.index(edge[0])]
                child_node_nx = nx_nodes[self.graph.nodes.index(edge[1])]
                nx_labels[(parent_node_nx, child_node_nx)] = labels[index]
            return nx_labels

        if not edges_labels and not nodes_labels:
            return

        bias = calculate_labels_bias(ax, longest_y_sequence)
        if nodes_labels:
            # Set labels for nodes
            labels_pos = deepcopy(pos)
            font_size = _get_scaled_font_size(longest_sequence, font_size_scale * 0.7)
            bbox = dict(alpha=0.9, color='w')
            for value in labels_pos.values():
                value[1] += bias

            nodes_nx_labels = match_labels_with_nx_nodes(nx_graph=nx_graph, labels=nodes_labels)
            nx.draw_networkx_labels(
                nx_graph, labels_pos,
                labels=nodes_nx_labels,
                font_color='black',
                font_size=font_size,
                bbox=bbox
            )

        if not edges_labels:
            return
        labels_pos_edges = deepcopy(pos)
        label_bias_y = 2 / 3 * bias
        if len(set([coord[1] for coord in pos.values()])) == 1 and len(list(pos.values())) > 2:
            for value in labels_pos_edges.values():
                value[1] += label_bias_y

        edges_nx_labels = match_labels_with_nx_edges(nx_graph=nx_graph, labels=edges_labels)
        # Set labels for edges
        for u, v, e in nx_graph.edges(data=True):
            if (u, v) not in edges_nx_labels:
                continue
            current_pos = labels_pos_edges
            if 'edge_center_position' in e:
                x, y = e['edge_center_position']
                plt.text(x, y, edges_nx_labels[(u, v)], bbox=bbox, fontsize=font_size)
            else:
                nx.draw_networkx_edge_labels(
                    nx_graph, current_pos, {(u, v): edges_nx_labels[(u, v)]},
                    label_pos=0.5, ax=ax,
                    font_color='black',
                    font_size=font_size,
                    rotate=False,
                    bbox=bbox
                )


def get_hierarchy_pos(graph: nx.DiGraph, max_line_length: int = 6) -> Tuple[Dict[Any, Tuple[float, float]], int]:
    """By default, returns 'networkx.multipartite_layout' positions based on 'hierarchy_level` from node data - the
     property must be set beforehand.
    If line of nodes reaches 'max_line_length', the result is the combination of 'networkx.multipartite_layout' and
    'networkx.spring_layout'.
    :param graph: the graph.
    :param max_line_length: the limit for common nodes horizontal or vertical line.
    """
    longest_path = nx.dag_longest_path(graph, weight=None)
    longest_sequence = len(longest_path)

    pos = nx.multipartite_layout(graph, subset_key='hierarchy_level')

    y_level_nodes_count = {}
    for x, _ in pos.values():
        y_level_nodes_count[x] = y_level_nodes_count.get(x, 0) + 1
        nodes_on_level = y_level_nodes_count[x]
        if nodes_on_level > longest_sequence:
            longest_sequence = nodes_on_level

    if longest_sequence > max_line_length:
        pos = {n: np.array(x_y) + (np.random.random(2) - 0.5) * 0.001 for n, x_y in pos.items()}
        pos = nx.spring_layout(graph, k=2, iterations=5, pos=pos, seed=42)

    return pos, longest_sequence


def remove_old_files_from_dir(dir_: Path, time_interval=datetime.timedelta(minutes=10)):
    for path in dir_.iterdir():
        if datetime.datetime.now() - datetime.datetime.fromtimestamp(path.stat().st_ctime) > time_interval:
            path.unlink()



from abc import ABC, abstractmethod
from enum import Enum
from os import PathLike
from typing import Dict, List, Optional, Sequence, Union, Tuple, TypeVar

from golem.core.dag.graph_node import GraphNode
from golem.visualisation.graph_viz import GraphVisualizer, NodeColorType

NodeType = TypeVar('NodeType', bound=GraphNode, covariant=False, contravariant=False)


class ReconnectType(Enum):
    """Defines allowed kinds of removals in Graph. Used by mutations."""
    none = 'none'  # do not reconnect predecessors
    single = 'single'  # reconnect a predecessor only if it's single
    all = 'all'  # reconnect all predecessors to all successors


class Graph(ABC):
    """Defines abstract graph interface that's required by graph optimisation process.
    """

    @abstractmethod
    def add_node(self, node: GraphNode):
        """Adds new node to the graph together with its parent nodes.

        Args:
            node: graph nodes
        """
        raise NotImplementedError()

    @abstractmethod
    def update_node(self, old_node: GraphNode, new_node: GraphNode):
        """Replaces ``old_node`` node with ``new_node``

        :param old_node: node to be replaced
        :param new_node: node to be placed instead
        """
        raise NotImplementedError()

    @abstractmethod
    def update_subtree(self, old_subtree: GraphNode, new_subtree: GraphNode):
        """Changes ``old_subtree`` subtree to ``new_subtree``

        Args:
            old_subtree: node and its subtree to be removed
            new_subtree: node and its subtree to be placed instead
        """
        raise NotImplementedError()

    @abstractmethod
    def delete_node(self, node: GraphNode, reconnect: ReconnectType = ReconnectType.single):
        """Removes ``node`` from the graph.
        If ``node`` has only one child, then connects all of the ``node`` parents to it.

        Args:
            node: node of the graph to be deleted
            reconnect: defines how to treat left edges between parents and children
        """
        raise NotImplementedError()

    @abstractmethod
    def delete_subtree(self, subtree: GraphNode):
        """Deletes given node with all its parents.
        Deletes all edges from removed nodes to remaining graph nodes

        Args:
            subtree: node to be deleted with all of its parents
                and their connections amongst the remaining graph nodes
        """
        raise NotImplementedError()

    @abstractmethod
    def node_children(self, node: GraphNode) -> Sequence[Optional[GraphNode]]:
        """Returns all children of the ``node``

        Args:
            node: for getting children from

        Returns: children of the ``node``
        """
        raise NotImplementedError()

    @abstractmethod
    def connect_nodes(self, node_parent: GraphNode, node_child: GraphNode):
        """Adds edge between ``parent`` and ``child``

        Args:
            node_parent: acts like parent in graph connection relations
            node_child:  acts like child in graph connection relations
        """
        raise NotImplementedError()

    @abstractmethod
    def disconnect_nodes(self, node_parent: GraphNode, node_child: GraphNode,
                         clean_up_leftovers: bool = False):
        """Removes an edge between two nodes

        Args:
            node_parent: where the removing edge comes out
            node_child: where the removing edge enters
            clean_up_leftovers: whether to remove the remaining invalid vertices with edges or not
        """
        raise NotImplementedError()

    @abstractmethod
    def get_edges(self) -> Sequence[Tuple[GraphNode, GraphNode]]:
        """Gets all available edges in this graph

        Returns:
            pairs of parent_node -> child_node
        """
        raise NotImplementedError()

    def get_nodes_by_name(self, name: str) -> List[GraphNode]:
        """Returns list of nodes with the required ``name``

        Args:
            name: name to filter by

        Returns:
            list: relevant nodes (empty if there are no such nodes)
        """

        appropriate_nodes = filter(lambda x: x.name == name, self.nodes)

        return list(appropriate_nodes)

    def get_node_by_uid(self, uid: str) -> Optional[GraphNode]:
        """Returns node with the required ``uid``

        Args:
            uid: uid of node to filter by

        Returns:
            Optional[Node]: relevant node (None if there is no such node)
        """

        appropriate_nodes = list(filter(lambda x: x.uid == uid, self.nodes))

        return appropriate_nodes[0] if appropriate_nodes else None

    @abstractmethod
    def __eq__(self, other_graph: 'Graph') -> bool:
        """Compares this graph with the ``other_graph``

        Args:
            other_graph: another graph

        Returns:
            is it equal to ``other_graph`` in terms of the graphs
        """
        raise NotImplementedError()

    def root_nodes(self) -> Sequence[GraphNode]:
        raise NotImplementedError()

    @property
    def root_node(self) -> Union[GraphNode, Sequence[GraphNode]]:
        """Gets the final layer node(s) of the graph

        Returns:
            the final layer node(s)
        """
        roots = self.root_nodes()
        if len(roots) == 1:
            return roots[0]
        return roots

    @property
    @abstractmethod
    def nodes(self) -> List[GraphNode]:
        """Return list of all graph nodes

        Returns:
            graph nodes
        """
        raise NotImplementedError()

    @nodes.setter
    @abstractmethod
    def nodes(self, new_nodes: List[GraphNode]):
        raise NotImplementedError()

    @property
    @abstractmethod
    def depth(self) -> int:
        """Gets this graph depth from its sink-node to its source-node

        Returns:
            length of a path from the root node to the farthest primary node
        """
        raise NotImplementedError()

    @property
    def length(self) -> int:
        """Return size of the graph (number of nodes)

        Returns:
            graph size
        """

        return len(self.nodes)

    def show(self, save_path: Optional[Union[PathLike, str]] = None, engine: Optional[str] = None,
             node_color: Optional[NodeColorType] = None, dpi: Optional[int] = None,
             node_size_scale: Optional[float] = None, font_size_scale: Optional[float] = None,
             edge_curvature_scale: Optional[float] = None,
             nodes_labels: Dict[int, str] = None, edges_labels: Dict[int, str] = None):
        """Visualizes graph or saves its picture to the specified ``path``

        Args:
            save_path: optional, save location of the graph visualization image.
            engine: engine to visualize the graph. Possible values: 'matplotlib', 'pyvis', 'graphviz'.
            node_color: color of nodes to use.
            node_size_scale: use to make node size bigger or lesser. Supported only for the engine 'matplotlib'.
            font_size_scale: use to make font size bigger or lesser. Supported only for the engine 'matplotlib'.
            edge_curvature_scale: use to make edges more or less curved. Supported only for the engine 'matplotlib'.
            dpi: DPI of the output image. Not supported for the engine 'pyvis'.
            nodes_labels: labels to display near nodes
            edges_labels: labels to display near edges
        """
        GraphVisualizer(graph=self)\
            .visualise(save_path=save_path, engine=engine, node_color=node_color, dpi=dpi,
                       node_size_scale=node_size_scale, font_size_scale=font_size_scale,
                       edge_curvature_scale=edge_curvature_scale,
                       nodes_labels=nodes_labels, edges_labels=edges_labels)

    @property
    def graph_description(self) -> Dict:
        """Return summary characteristics of the graph

        Returns:
            dict: containing information about the graph
        """
        return {
            'depth': self.depth,
            'length': self.length,
            'nodes': self.nodes,
        }

    @property
    def descriptive_id(self) -> str:
        """Returns human-readable identifier of the graph.

        Returns:
            str: text description of the content in the node and its parameters
        """
        return self.root_node.descriptive_id

    def __str__(self):
        return str(self.graph_description)

    def __repr__(self):
        return self.__str__()

    def __len__(self):
        return self.length



from typing import Dict, Tuple, Callable, List


class SearchSpace:
    """
    Args:
        search_space: dictionary with parameters and their search_space
            {'operation_name': {'param_name': (hyperopt distribution function, [sampling scope]), ...}, ...},
            e.g. ``{'operation_name': {'param1': (hp.uniformint, [2, 21]), ...}, ..}
    """
    def __init__(self, search_space: Dict[str, Dict[str, Tuple[Callable, List]]]):
        self.parameters_per_operation = search_space

    def get_operation_parameter_range(self, operation_name: str, parameter_name: str = None, label: str = 'default'):
        """
        Method return hyperopt object with search_space from search_space dictionary
        If parameter name is not defined - return all available operations

        Args:
            operation_name: name of the operation
            parameter_name: name of hyperparameter of particular operation
            label: label to assign in hyperopt pyll

        Returns:
            dictionary with appropriate range
        """

        # Get available parameters for current operation
        operation_parameters = self.parameters_per_operation.get(operation_name)

        if operation_parameters is not None:
            # If there are not parameter_name - return list with all parameters
            if parameter_name is None:
                return list(operation_parameters)
            else:
                hyperopt_tuple = operation_parameters.get(parameter_name)
                return hyperopt_tuple[0](label, *hyperopt_tuple[1])
        else:
            return None

    def get_node_params(self, node_id, operation_name):
        """
        Method for forming dictionary with hyperparameters for considering
        operation as a part of the whole graph

        :param node_id: number of node in graph.nodes list
        :param operation_name: name of operation in the node

        :return params_dict: dictionary-like structure with labeled hyperparameters
        and their range per operation
        """

        # Get available parameters for operation
        params_list = self.get_operation_parameter_range(operation_name)

        if params_list is None:
            params_dict = None
        else:
            params_dict = {}
            for parameter_name in params_list:
                node_op_parameter_name = get_node_operation_parameter_label(node_id, operation_name, parameter_name)

                # For operation get range where search can be done
                space = self.get_operation_parameter_range(operation_name=operation_name,
                                                           parameter_name=parameter_name,
                                                           label=node_op_parameter_name)

                params_dict.update({node_op_parameter_name: space})

        return params_dict


def get_node_operation_parameter_label(node_id: int, operation_name: str, parameter_name: str) -> str:
    # Name with operation and parameter
    op_parameter_name = ''.join((operation_name, ' | ', parameter_name))

    # Name with node id || operation | parameter
    node_op_parameter_name = ''.join((str(node_id), ' || ', op_parameter_name))
    return node_op_parameter_name


def convert_params(params):
    """
    Function removes labels from dictionary with operations

    :param params: labeled parameters
    :return new_params: dictionary without labels of node_id and operation_name
    """

    new_params = {}
    for operation_parameter, value in params.items():
        # Remove right part of the parameter name
        parameter_name = operation_parameter.split(' | ')[-1]

        if value is not None:
            new_params.update({parameter_name: value})

    return new_params



from copy import deepcopy
from random import choice
from typing import Sequence, Callable, Union, Any

from golem.core.constants import MAX_GRAPH_GEN_ATTEMPTS
from golem.core.dag.graph import Graph
from golem.core.optimisers.genetic.gp_params import GPAlgorithmParameters
from golem.core.optimisers.genetic.operators.crossover import Crossover
from golem.core.optimisers.genetic.operators.elitism import Elitism
from golem.core.optimisers.genetic.operators.inheritance import Inheritance
from golem.core.optimisers.genetic.operators.mutation import Mutation
from golem.core.optimisers.genetic.operators.operator import PopulationT, EvaluationOperator
from golem.core.optimisers.genetic.operators.regularization import Regularization
from golem.core.optimisers.genetic.operators.selection import Selection
from golem.core.optimisers.genetic.parameters.graph_depth import AdaptiveGraphDepth
from golem.core.optimisers.genetic.parameters.operators_prob import init_adaptive_operators_prob
from golem.core.optimisers.genetic.parameters.population_size import init_adaptive_pop_size, PopulationSize
from golem.core.optimisers.optimization_parameters import GraphRequirements
from golem.core.optimisers.objective.objective import Objective
from golem.core.optimisers.opt_history_objects.individual import Individual
from golem.core.optimisers.optimizer import GraphGenerationParams
from golem.core.optimisers.populational_optimizer import PopulationalOptimizer, EvaluationAttemptsError


EVALUATION_ATTEMPTS_NUMBER = 5


class EvoGraphOptimizer(PopulationalOptimizer):
    """
    Multi-objective evolutionary graph optimizer named GPComp
    """

    def __init__(self,
                 objective: Objective,
                 initial_graphs: Sequence[Union[Graph, Any]],
                 requirements: GraphRequirements,
                 graph_generation_params: GraphGenerationParams,
                 graph_optimizer_params: GPAlgorithmParameters):
        super().__init__(objective, initial_graphs, requirements, graph_generation_params, graph_optimizer_params)
        # Define genetic operators
        self.regularization = Regularization(graph_optimizer_params, graph_generation_params)
        self.selection = Selection(graph_optimizer_params)
        self.crossover = Crossover(graph_optimizer_params, requirements, graph_generation_params)
        self.mutation = Mutation(graph_optimizer_params, requirements, graph_generation_params)
        self.inheritance = Inheritance(graph_optimizer_params, self.selection)
        self.elitism = Elitism(graph_optimizer_params)
        self.operators = [self.regularization, self.selection, self.crossover,
                          self.mutation, self.inheritance, self.elitism]

        # Define adaptive parameters
        self._pop_size: PopulationSize = init_adaptive_pop_size(graph_optimizer_params, self.generations)
        self._operators_prob = init_adaptive_operators_prob(graph_optimizer_params)
        self._graph_depth = AdaptiveGraphDepth(self.generations,
                                               start_depth=requirements.start_depth,
                                               max_depth=requirements.max_depth,
                                               max_stagnation_gens=graph_optimizer_params.adaptive_depth_max_stagnation,
                                               adaptive=graph_optimizer_params.adaptive_depth)

        # Define initial parameters
        self.requirements.max_depth = self._graph_depth.initial
        self.graph_optimizer_params.pop_size = self._pop_size.initial
        self.initial_individuals = [Individual(graph, metadata=requirements.static_individual_metadata)
                                    for graph in self.initial_graphs]

    def _initial_population(self, evaluator: EvaluationOperator):
        """ Initializes the initial population """
        # Adding of initial assumptions to history as zero generation
        self._update_population(evaluator(self.initial_individuals), 'initial_assumptions')

        if len(self.initial_individuals) < self.graph_optimizer_params.pop_size:
            self.initial_individuals = self._extend_population(self.initial_individuals)
            # Adding of extended population to history
            self._update_population(evaluator(self.initial_individuals), 'extended_initial_assumptions')

    def _extend_population(self, initial_individuals: PopulationT) -> PopulationT:
        initial_individuals = list(initial_individuals)
        initial_graphs = [ind.graph for ind in initial_individuals]
        initial_req = deepcopy(self.requirements)
        initial_req.mutation_prob = 1
        self.mutation.update_requirements(requirements=initial_req)

        for iter_num in range(MAX_GRAPH_GEN_ATTEMPTS):
            if len(initial_individuals) == self.graph_optimizer_params.pop_size:
                break
            new_ind = self.mutation(choice(self.initial_individuals))
            new_graph = new_ind.graph
            if new_graph not in initial_graphs and self.graph_generation_params.verifier(new_graph):
                initial_individuals.append(new_ind)
                initial_graphs.append(new_graph)
        else:
            self.log.warning(f'Exceeded max number of attempts for extending initial graphs, stopping.'
                             f'Current size {len(self.initial_individuals)} '
                             f'instead of {self.graph_optimizer_params.pop_size} graphs.')

        self.mutation.update_requirements(requirements=self.requirements)
        return initial_individuals

    def _evolve_population(self, evaluator: EvaluationOperator) -> PopulationT:
        """ Method realizing full evolution cycle """
        self._update_requirements()

        individuals_to_select = self.regularization(self.population, evaluator)
        selected_individuals = self.selection(individuals_to_select)
        new_population = self._spawn_evaluated_population(selected_individuals, evaluator)
        new_population = self.inheritance(self.population, new_population)
        new_population = self.elitism(self.generations.best_individuals, new_population)

        return new_population

    def _update_requirements(self):
        if not self.generations.is_any_improved:
            self.graph_optimizer_params.mutation_prob, self.graph_optimizer_params.crossover_prob = \
                self._operators_prob.next(self.population)
        self.graph_optimizer_params.pop_size = self._pop_size.next(self.population)
        self.requirements.max_depth = self._graph_depth.next()
        self.log.info(
            f'Next population size: {self.graph_optimizer_params.pop_size}; '
            f'max graph depth: {self.requirements.max_depth}')

        # update requirements in operators
        for operator in self.operators:
            operator.update_requirements(self.graph_optimizer_params, self.requirements)

    def _spawn_evaluated_population(self, selected_individuals: PopulationT, evaluator: Callable) -> PopulationT:
        """ Reproduce and evaluate new population. If at least one of received individuals can not be evaluated then
        mutate and evaluate selected individuals until a new population is obtained
        or the number of attempts is exceeded """
        for i in range(EVALUATION_ATTEMPTS_NUMBER):
            new_population = self.crossover(selected_individuals)
            new_population = self.mutation(new_population)
            new_population = evaluator(new_population)
            if new_population:
                return new_population
        else:
            # Could not generate valid population; raise an error
            raise EvaluationAttemptsError()



from copy import deepcopy
from typing import Optional, Dict, Any, Iterable

import networkx as nx
import numpy as np

from golem.core.adapter import BaseOptimizationAdapter
from golem.core.dag.graph_node import GraphNode
from golem.core.optimisers.graph import OptGraph, OptNode


class BaseNetworkxAdapter(BaseOptimizationAdapter[nx.DiGraph]):
    """Base class for adaptation of networkx.DiGraph to optimization graph.
    Allows to use NetworkX directed graphs with our optimizers.

    For custom networkx adapters overwrite methods responsible for
    transformation of single nodes (`_node_adapt` & `_node_restore`).
    """

    def __init__(self):
        super().__init__(base_graph_class=nx.DiGraph)

    def _node_restore(self, node: GraphNode) -> Dict:
        """Transforms GraphNode to dict of NetworkX node attributes.
        Override for custom behavior."""
        if hasattr(node, 'content'):
            return deepcopy(node.content)
        else:
            return {}

    def _node_adapt(self, data: Dict) -> OptNode:
        """Transforms a dict of NetworkX node attributes to GraphNode.
        Override for custom behavior."""
        return OptNode(content=deepcopy(data))

    def _adapt(self, adaptee: nx.DiGraph) -> OptGraph:
        mapped_nodes = {}

        def map_predecessors(node_id) -> Iterable[OptNode]:
            for pred_id in adaptee.predecessors(node_id):
                yield mapped_nodes[pred_id]

        # map nodes
        for node_id, node_data in adaptee.nodes.items():
            # transform node
            node = self._node_adapt(node_data)
            mapped_nodes[node_id] = node

        # map parent nodes
        for node_id, node in mapped_nodes.items():
            # append its parent edges
            node.nodes_from = map_predecessors(node_id)

        return OptGraph(mapped_nodes.values())

    def _restore(self, opt_graph: OptGraph, metadata: Optional[Dict[str, Any]] = None) -> nx.DiGraph:
        nx_graph = nx.DiGraph()
        nx_node_data = {}

        # add nodes
        for node in opt_graph.nodes:
            nx_node_data[node.uid] = self._node_restore(node)
            nx_graph.add_node(node.uid)

        # add edges
        for node in opt_graph.nodes:
            for parent in node.nodes_from:
                nx_graph.add_edge(parent.uid, node.uid)

        # add nodes ad labels
        nx.set_node_attributes(nx_graph, nx_node_data)

        return nx_graph


_NX_NODE_KEY = 'data'


class DumbNetworkxAdapter(BaseNetworkxAdapter):
    """Simple version of networkx adapter that just stores
    `OptNode` as an attribute of NetworkX graph node."""

    def _node_restore(self, node: GraphNode) -> Dict:
        return {_NX_NODE_KEY: node}

    def _node_adapt(self, data: Dict) -> OptNode:
        return data[_NX_NODE_KEY]



from typing import Union, Sequence, List, TYPE_CHECKING, Callable

if TYPE_CHECKING:
    from golem.core.dag.graph import Graph
    from golem.core.dag.graph_node import GraphNode


def distance_to_root_level(graph: 'Graph', node: 'GraphNode') -> int:
    """Gets distance to the final output node

    Args:
        graph: graph for finding the distance
        node: search starting point

    Return:
        int: distance to root level
    """

    def recursive_child_height(parent_node: 'GraphNode') -> int:
        node_child = graph.node_children(parent_node)
        if node_child:
            height = recursive_child_height(node_child[0]) + 1
            return height
        return 0

    height = recursive_child_height(node)
    return height


def distance_to_primary_level(node: 'GraphNode') -> int:
    return node_depth(node) - 1


def nodes_from_layer(graph: 'Graph', layer_number: int) -> Sequence['GraphNode']:
    """Gets all the nodes from the chosen layer up to the surface

    Args:
        graph: graph with nodes
        layer_number: max height of diving

    Returns:
        all nodes from the surface to the ``layer_number`` layer
    """

    def get_nodes(roots: Sequence['GraphNode'], current_height: int) -> Sequence['GraphNode']:
        """Gets all the parent nodes of ``roots``

        :param roots: nodes to get all subnodes from
        :param current_height: current diving step depth

        :return: all parent nodes of ``roots`` in one sequence:69
        """
        nodes = []
        if current_height == layer_number:
            nodes.extend(roots)
        else:
            for root in roots:
                nodes.extend(get_nodes(root.nodes_from, current_height + 1))
        return nodes

    nodes = get_nodes(graph.root_nodes(), current_height=0)
    return nodes


def ordered_subnodes_hierarchy(node: 'GraphNode') -> List['GraphNode']:
    """Gets hierarchical subnodes representation of the graph starting from the bounded node

    Returns:
        List['GraphNode']: hierarchical subnodes list starting from the bounded node
    """
    started = {node}
    visited = set()

    def subtree_impl(node):
        nodes = [node]
        for parent in node.nodes_from:
            if parent in visited:
                continue
            elif parent in started:
                raise ValueError('Can not build ordered node hierarchy: graph has cycle')
            started.add(parent)
            nodes.extend(subtree_impl(parent))
            visited.add(parent)
        return nodes

    return subtree_impl(node)


def node_depth(node: 'GraphNode') -> int:
    """Gets this graph depth from the provided ``node`` to the graph source node

    Args:
        node: where to start diving from

    Returns:
        int: length of a path from the provided ``node`` to the farthest primary node
    """
    if not node.nodes_from:
        return 1
    else:
        return 1 + max(node_depth(next_node) for next_node in node.nodes_from)


def map_dag_nodes(transform: Callable, nodes: Sequence) -> Sequence:
    """Maps nodes in dfs-order while respecting node edges.

    Args:
        transform: node transform function (maps node to node)
        nodes: sequence of nodes for mapping

    Returns:
        Sequence: sequence of transformed links with preserved relations
    """
    mapped_nodes = {}

    def map_impl(node):
        already_mapped = mapped_nodes.get(id(node))
        if already_mapped:
            return already_mapped
        # map node itself
        mapped_node = transform(node)
        # remember it to avoid recursion
        mapped_nodes[id(node)] = mapped_node
        # map its children
        mapped_node.nodes_from = list(map(map_impl, node.nodes_from))
        return mapped_node

    return list(map(map_impl, nodes))


def graph_structure(graph: 'Graph') -> str:
    """ Returns structural information about the graph - names and parameters of graph nodes.
    Represents graph info in easily readable way.

    Returns:
        str: graph structure
    """
    return '\n'.join([str(graph), *(f'{node.name} - {node.parameters}' for node in graph.nodes)])



import collections.abc

from abc import ABC, abstractmethod
from copy import deepcopy
from enum import Enum
from typing import Callable, Container, Generic, Iterable, Iterator, List, Optional, Sequence, Sized, TypeVar, Union

T = TypeVar('T')


class UniqueList(list, Generic[T]):
    """
    Simple list that maintains uniqueness of elements.
    In comparison to 'set': preserves list interface and element ordering.

    But this class is better to use only for short lists because of linear complexity of uniqueness lookup.
    Behaves optimally for <= 4 items, good enough for <= 20 items.

    List addition and multiplication are not overloaded and return a standard list.
    But addition with assignment (+=) behaves as UniqueList.extend.

    :param iterable: list of elements to turn into distinct
    """

    def __init__(self, iterable: Optional[Iterable[T]] = None):
        iterable = iterable or ()
        super().__init__(dict.fromkeys(iterable).keys())  # preserves order and uniqueness in the newly created list

    def append(self, value: T):
        """
        Adds ``value`` to the end of the list in case it is not present there

        :param value: will be added to the list or ignored if it's already there
        """
        if value not in super().__iter__():
            super().append(value)

    def extend(self, iterable: Iterable[T]):
        """
        Extends current list with the ``iterable`` elements that aren't present in the list yet

        :param iterable: sequence of elements to be added to the list in case they aren't there
        """
        impl = super()
        impl.extend(element for element in iterable
                    if element not in impl.__iter__())

    def insert(self, index: int, value: T):
        """
        Inserts specified ``value`` to the provided ``index`` in the list in case it's unique value

        :param index: position for inserting into the list
        :param value: will be inserted at the specified ``index`` if it's unique
        """
        if value not in super().__iter__():
            super().insert(index, value)

    def __setitem__(self, key: Union[int, slice], value: Union[T, Iterable[T]]):
        """
        Sets specified ``value`` at the specified index or slice if it's unique value

        :param key: for pointing to the elements from the list
        :param value: element(s) to be set by the specified ``key``
        """
        if value not in super().__iter__():
            super().__setitem__(key, value)

    def __iadd__(self, other: Iterable[T]) -> 'UniqueList':
        """
        Extends current list with the ``iterable`` elements that aren't present in the list yet

        :param other: sequence of elements to be added to the list in case they aren't there

        :return: this class instance
        """
        self.extend(other)
        return self


def remove_items(collection: List[T], items_to_remove: Container[T]):
    """
    Removes all specified items from the list. Modifies original collection

    :param collection: list of elements to be filtered
    :param items_to_remove: filter list of unwanted elements

    :return: modified ``collection`` parameter
    """
    if collection:
        collection[:] = [item for item in collection if item not in items_to_remove]
    return collection


def are_same_length(collections: Iterable[Sized]) -> bool:
    """
    Checks if all arguments have the same length

    :param collections: collection of collections

    :return: does collections inside ``collections`` have the same length
    """
    it = collections.__iter__()
    first = next(it, None)
    if first is not None:
        first = len(first)
        for elem in it:
            if len(elem) != first:
                return False
    return True


def ensure_wrapped_in_sequence(
        obj: Optional[Union[T, Iterable[T]]],
        sequence_factory: Callable[[Iterable[T]], Sequence[T]] = list
) -> Optional[Sequence[T]]:
    """
    Makes sure given ``obj`` is of type that acts like sequence and converts ``obj`` to it otherwise

    :param obj: any object to be ensured or wrapped into sequence type
    :param sequence_factory: sequence factory for wrapping ``obj`` into in case it is not of any sequence type

    :return: the same object if it's of type sequence (or None)
        or ``obj`` wrapped in type provided by ``sequence_factory``
    """
    if obj is None:
        return obj
    elif isinstance(obj, str) or not isinstance(obj, collections.abc.Iterable):
        return sequence_factory([obj])
    elif not isinstance(obj, collections.abc.Sequence) or not isinstance(obj, sequence_factory):
        return sequence_factory(obj)
    return obj


class ComparableEnum(Enum):
    """
    The Enum implementation that allows to avoid the multi-module enum comparison problem
    (https://stackoverflow.com/questions/26589805/python-enums-across-modules)
    """

    def __eq__(self, other: 'Enum'):
        """
        Compares this enum with the ``other_graph``

        :param other: another enum

        :return: is it equal to ``other`` in terms of the string representation
        """
        return str(self) == str(other)

    def __hash__(self):
        """
        Gets hashcode of this enum

        :return: hashcode of string representation of this enum
        """
        return hash(str(self))


class Copyable:
    """Provides default implementations for `copy` & `deepcopy`."""

    def __copy__(self):
        cls = self.__class__
        result = cls.__new__(cls)
        result.__dict__.update(self.__dict__)
        return result

    def __deepcopy__(self, memo=None):
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            setattr(result, k, deepcopy(v, memo))
        return result


class Comparable(ABC):
    @abstractmethod
    def __eq__(self, other: 'Comparable') -> bool:
        """
        Compares this object of :class:`Comparable` interface with the ``other`` comparable

        :param other: other object implementing :class:`Comparable` interface

        :return: is it equal to ``other`` in terms of the comparables
        """
        raise NotImplementedError()

    @abstractmethod
    def __lt__(self, other: 'Comparable') -> bool:
        """
        Compares this object of :class:`Comparable` interface with the ``other`` comparable

        :param other: other object implementing :class:`Comparable` interface

        :return: is less than ``other`` in terms of the comparables
        """
        raise NotImplementedError()

    def __ne__(self, other: 'Comparable') -> bool:
        """
        Compares this object of :class:`Comparable` interface with the ``other`` comparable

        :param other: other object implementing :class:`Comparable` interface

        :return: is it not equal to ``other`` in terms of the comparables
        """
        return not self.__eq__(other)

    def __le__(self, other) -> bool:
        """
        Compares this object of :class:`Comparable` interface with the ``other`` comparable

        :param other: other object implementing :class:`Comparable` interface

        :return: is less than or equal to ``other`` in terms of the comparables
        """
        return self.__lt__(other) or self.__eq__(other)

    def __gt__(self, other) -> bool:
        """
        Compares this object of :class:`Comparable` interface with the ``other`` comparable

        :param other: other object implementing :class:`Comparable` interface

        :return: is greater than ``other`` in terms of the comparables
        """
        return not self.__le__(other)

    def __ge__(self, other) -> bool:
        """
        Compares this object of :class:`Comparable` interface with the ``other`` comparable

        :param other: other object implementing :class:`Comparable` interface

        :return: is greater than or equal to ``other`` in terms of the comparables
        """
        return not self.__lt__(other)


class BidirectionalIterator(Iterator[T]):
    @abstractmethod
    def has_prev(self) -> bool:
        """
        Checks if this iterator has implemented previous item getter

        :return: whether this iterator has previous item getter
        """
        raise NotImplementedError()

    @abstractmethod
    def has_next(self) -> bool:
        """
        Checks if this iterator has implemented next item getter

        :return: whether this iterator has next item getter
        """
        raise NotImplementedError()

    @abstractmethod
    def next(self) -> T:
        """
        Gets next item of this iterator

        :return: next item
        """
        raise NotImplementedError()

    @abstractmethod
    def prev(self) -> T:
        """
        Gets previous item of this iterator

        :return: previous item
        """
        raise NotImplementedError()

    @abstractmethod
    def current(self) -> T:
        """
        Get current value pointed to by the iterator

        :return: current item
        """
        raise NotImplementedError()

    def __next__(self):
        """
        Gets next item of this iterator

        :raises StopIteration: the end of this iterator

        :return: next item
        """
        if self.has_next():
            return self.next()
        raise StopIteration()

    def __iter__(self):
        """
        Returns this iterator

        :return: this iterator
        """
        return self



from dataclasses import dataclass
from typing import Sequence, Union, Any

from golem.core.optimisers.genetic.operators.base_mutations import MutationStrengthEnum, MutationTypesEnum, \
    simple_mutation_set
from golem.core.optimisers.optimizer import AlgorithmParameters
from golem.core.optimisers.genetic.operators.crossover import CrossoverTypesEnum
from golem.core.optimisers.genetic.operators.elitism import ElitismTypesEnum
from golem.core.optimisers.genetic.operators.inheritance import GeneticSchemeTypesEnum
from golem.core.optimisers.genetic.operators.regularization import RegularizationTypesEnum
from golem.core.optimisers.genetic.operators.selection import SelectionTypesEnum


@dataclass
class GPAlgorithmParameters(AlgorithmParameters):
    """
    Defines parameters of evolutionary operators and the algorithm of genetic optimizer.

    :param crossover_prob: crossover probability (the chance that two chromosomes exchange some of their parts)
    :param mutation_prob: mutation probability
    :param static_mutation_prob: probability of applying same mutation to graph in a cycle of mutations
    :param max_num_of_operator_attempts: max number of unsuccessful operator (mutation/crossover)
    attempts before continuing
    :param mutation_strength: strength of mutation in tree (using in certain mutation types)
    :param min_pop_size_with_elitism: minimal population size with which elitism is applicable

    :param selection_types: Sequence of selection operators types
    :param crossover_types: Sequence of crossover operators types
    :param mutation_types: Sequence of mutation operators types
    :param elitism_type: type of elitism operator evolution

    :param regularization_type: type of regularization operator

    Regularization attempts to cut off the subtrees of the graph. If the truncated graph
    is not worse than the original, then it enters the new generation as a simpler solution.
    Regularization is not used by default, it must be explicitly enabled.

    :param genetic_scheme_type: type of genetic evolutionary scheme

    The `generational` scheme is a standard scheme of the evolutionary algorithm.
    It specifies that at each iteration the entire generation is updated.

    In the `steady_state` scheme at each iteration only one individual is updated.

    The `parameter_free` scheme is an adaptive variation of the `generational` scheme.
    It specifies that the population size and the probability of mutation and crossover
    change depending on the success of convergence. If there are no improvements in fitness,
    then the size and the probabilities increase. When fitness improves, the size and the
    probabilities decrease. That is, the algorithm choose a more stable and conservative
    mode when optimization seems to converge.
    """

    crossover_prob: float = 0.8
    mutation_prob: float = 0.8
    static_mutation_prob: float = 0.7
    max_num_of_operator_attempts: int = 100
    mutation_strength: MutationStrengthEnum = MutationStrengthEnum.mean
    min_pop_size_with_elitism: int = 5

    selection_types: Sequence[SelectionTypesEnum] = \
        (SelectionTypesEnum.tournament,)
    crossover_types: Sequence[Union[CrossoverTypesEnum, Any]] = \
        (CrossoverTypesEnum.one_point,)
    mutation_types: Sequence[Union[MutationTypesEnum, Any]] = simple_mutation_set
    elitism_type: ElitismTypesEnum = ElitismTypesEnum.keep_n_best
    regularization_type: RegularizationTypesEnum = RegularizationTypesEnum.none
    genetic_scheme_type: GeneticSchemeTypesEnum = GeneticSchemeTypesEnum.generational

    def __post_init__(self):
        if self.multi_objective:
            self.selection_types = (SelectionTypesEnum.spea2,)
            # TODO add possibility of using regularization in MO alg
            self.regularization_type = RegularizationTypesEnum.none



from abc import abstractmethod
from copy import deepcopy
from datetime import timedelta
from typing import Callable, TypeVar, Generic

import numpy as np
from hyperopt.early_stop import no_progress_loss

from golem.core.adapter import BaseOptimizationAdapter
from golem.core.adapter.adapter import IdentityAdapter
from golem.core.constants import MAX_TUNING_METRIC_VALUE
from golem.core.dag.graph_utils import graph_structure
from golem.core.log import default_log
from golem.core.optimisers.graph import OptGraph
from golem.core.optimisers.objective import ObjectiveEvaluate
from golem.core.optimisers.timer import Timer
from golem.core.tuning.search_space import SearchSpace

DomainGraphForTune = TypeVar('DomainGraphForTune')


class HyperoptTuner(Generic[DomainGraphForTune]):
    """
    Base class for hyperparameters optimization based on hyperopt library
    
    Args:
      objective_evaluate: objective to optimize
      adapter: the function for processing of external object that should be optimized
      iterations: max number of iterations
      timeout: max time for tuning
      search_space: SearchSpace instance
      algo: algorithm for hyperparameters optimization with signature similar to :obj:`hyperopt.tse.suggest`
      n_jobs: num of ``n_jobs`` for parallelization (``-1`` for use all cpu's)
      deviation: required improvement (in percent) of a metric to return tuned graph.
        By default, ``deviation=0.05``, which means that tuned graph will be returned
        if it's metric will be at least 0.05% better than the initial.
    """

    def __init__(self, objective_evaluate: ObjectiveEvaluate,
                 search_space: SearchSpace,
                 adapter: BaseOptimizationAdapter = None,
                 iterations=100, early_stopping_rounds=None,
                 timeout: timedelta = timedelta(minutes=5),
                 algo: Callable = None,
                 n_jobs: int = -1,
                 deviation: float = 0.05):
        self.iterations = iterations
        self.adapter = adapter or IdentityAdapter()
        iteration_stop_count = early_stopping_rounds or max(100, int(np.sqrt(iterations) * 10))
        self.early_stop_fn = no_progress_loss(iteration_stop_count=iteration_stop_count)
        self.max_seconds = int(timeout.seconds) if timeout is not None else None
        self.init_graph = None
        self.init_metric = None
        self.obtained_metric = None
        self.n_jobs = n_jobs
        objective_evaluate.eval_n_jobs = self.n_jobs
        self.objective_evaluate = self.adapter.adapt_func(objective_evaluate.evaluate)
        self._default_metric_value = MAX_TUNING_METRIC_VALUE
        self.search_space = search_space
        self.algo = algo
        self.deviation = deviation
        self.was_tuned = False
        self.log = default_log(self)

    @abstractmethod
    def tune(self, graph: DomainGraphForTune) -> DomainGraphForTune:
        """
        Function for hyperparameters tuning on the graph

        Args:
          graph: domain graph for which hyperparameters tuning is needed

        Returns:
          Graph with optimized hyperparameters
        """
        raise NotImplementedError()

    def get_metric_value(self, graph: OptGraph) -> float:
        """
        Method calculates metric for algorithm validation
        
        Args:
          graph: Graph to evaluate

        Returns:
          value of loss function
        """
        graph_fitness = self.objective_evaluate(graph)
        metric_value = graph_fitness.value
        if not graph_fitness.valid:
            return self._default_metric_value
        return metric_value

    def init_check(self, graph: OptGraph) -> None:
        """
        Method get metric on validation set before start optimization

        Args:
          graph: graph to calculate objective
        """
        self.log.info('Hyperparameters optimization start: estimation of metric for initial graph')

        # Train graph
        self.init_graph = deepcopy(graph)

        self.init_metric = self.get_metric_value(graph=self.init_graph)
        self.log.message(f'Initial graph: {graph_structure(self.init_graph)} \n'
                         f'Initial metric: {abs(self.init_metric):.3f}')

    def final_check(self, tuned_graph: OptGraph) -> OptGraph:
        """
        Method propose final quality check after optimization process

        Args:
          tuned_graph: Tuned graph to calculate objective
        """

        self.obtained_metric = self.get_metric_value(graph=tuned_graph)

        if self.obtained_metric == self._default_metric_value:
            self.obtained_metric = None

        self.log.info('Hyperparameters optimization finished')

        prefix_tuned_phrase = 'Return tuned graph due to the fact that obtained metric'
        prefix_init_phrase = 'Return init graph due to the fact that obtained metric'

        # 0.05% deviation is acceptable
        deviation_value = (self.init_metric / 100.0) * self.deviation
        init_metric = self.init_metric + deviation_value * (-np.sign(self.init_metric))
        if self.obtained_metric is None:
            self.log.info(f'{prefix_init_phrase} is None. Initial metric is {abs(init_metric):.3f}')
            final_graph = self.init_graph
            final_metric = self.init_metric
        elif self.obtained_metric <= init_metric:
            self.log.info(f'{prefix_tuned_phrase} {abs(self.obtained_metric):.3f} equal or '
                          f'better than initial (+ {self.deviation}% deviation) {abs(init_metric):.3f}')
            final_graph = tuned_graph
            final_metric = self.obtained_metric
        else:
            self.log.info(f'{prefix_init_phrase} {abs(self.obtained_metric):.3f} '
                          f'worse than initial (+ {self.deviation}% deviation) {abs(init_metric):.3f}')
            final_graph = self.init_graph
            final_metric = self.init_metric
        self.log.message(f'Final graph: {graph_structure(final_graph)}')
        if final_metric is not None:
            self.log.message(f'Final metric: {abs(final_metric):.3f}')
        else:
            self.log.message(f'Final metric is None')
        return final_graph

    def _update_remaining_time(self, tuner_timer: Timer):
        self.max_seconds = self.max_seconds - tuner_timer.minutes_from_start * 60

    def _stop_tuning_with_message(self, message: str):
        self.log.message(message)
        self.obtained_metric = self.init_metric



from datetime import datetime, timedelta
from functools import partial
from io import StringIO
from itertools import product
from pathlib import Path
from typing import Sequence, Type, Callable, Optional

from examples.synthetic_graph_evolution.generators import generate_labeled_graph, graph_kinds
from examples.synthetic_graph_evolution.utils import draw_graphs_subplots
from golem.core.adapter.nx_adapter import BaseNetworkxAdapter
from golem.core.optimisers.genetic.gp_optimizer import EvoGraphOptimizer
from golem.core.optimisers.optimizer import GraphOptimizer
from golem.metrics.edit_distance import get_edit_dist_metric, matrix_edit_dist
from golem.metrics.graph_metrics import *


def get_all_quality_metrics(target_graph):
    quality_metrics = {
        'edit_distance': get_edit_dist_metric(target_graph),
        'matrix_edit_dist': partial(matrix_edit_dist, target_graph),
        'sp_adj': partial(spectral_dist, target_graph, kind='adjacency'),
        'sp_lapl': partial(spectral_dist, target_graph, kind='laplacian'),
        'sp_lapl_norm': partial(spectral_dist, target_graph, kind='laplacian_norm'),
        'graph_size': partial(size_diff, target_graph),
        'degree_dist_mmd': partial(degree_distance_kernel, target_graph),
        'degree_dist': partial(degree_distance, target_graph),
    }
    return quality_metrics


def run_experiments(optimizer_setup: Callable,
                    optimizer_cls: Type[GraphOptimizer] = EvoGraphOptimizer,
                    node_types: Optional[Sequence[str]] = None,
                    graph_names: Sequence[str] = graph_kinds,
                    graph_sizes: Sequence[int] = (30, 100, 300),
                    num_trials: int = 1,
                    trial_timeout: Optional[int] = None,
                    trial_iterations: Optional[int] = None,
                    visualize: bool = False,
                    ):
    Path("results").mkdir(exist_ok=True)
    log = StringIO()
    if not node_types:
        node_types = ['X']
    for graph_name, num_nodes in product(graph_names, graph_sizes):
        experiment_id = f'Experiment [graph={graph_name} graph_size={num_nodes}]'
        trial_results = []
        for i in range(num_trials):
            start_time = datetime.now()
            print(f'\nTrial #{i} of {experiment_id} started at {start_time}', file=log)

            # Generate random target graph and run the optimizer
            target_graph = generate_labeled_graph(graph_name, num_nodes, node_types)
            # Run optimizer setup
            optimizer, objective = optimizer_setup(target_graph,
                                                   optimizer_cls=optimizer_cls,
                                                   node_types=node_types,
                                                   timeout=timedelta(minutes=trial_timeout),
                                                   num_iterations=trial_iterations)
            found_graphs = optimizer.optimise(objective)
            found_graph = found_graphs[0] if isinstance(found_graphs, Sequence) else found_graphs
            history = optimizer.history

            trial_results.extend(history.final_choices)
            found_nx_graph = BaseNetworkxAdapter().restore(found_graph)

            duration = datetime.now() - start_time
            print(f'Trial #{i} finished, spent time: {duration}', file=log)
            print('target graph stats: ', nxgraph_stats(target_graph), file=log)
            print('found graph stats: ', nxgraph_stats(found_nx_graph), file=log)
            if visualize:
                draw_graphs_subplots(target_graph, found_nx_graph,
                                     titles=['Target Graph', 'Found Graph'])
                history.show.fitness_line()
            history.save(f'./results/hist_{graph_name}_n{num_nodes}_trial{i}.json')

        # Compute mean & std for metrics of trials
        ff = objective.format_fitness
        trial_metrics = np.array([ind.fitness.values for ind in trial_results])
        trial_metrics_mean = trial_metrics.mean(axis=0)
        trial_metrics_std = trial_metrics.std(axis=0)
        print(f'{experiment_id} finished with metrics:\n'
              f'mean={ff(trial_metrics_mean)}\n'
              f' std={ff(trial_metrics_std)}',
              file=log)
        print(log.getvalue())
    return log.getvalue()


def run_trial(target_graph: nx.DiGraph,
              optimizer_setup: Callable,
              optimizer_cls: Type[GraphOptimizer] = EvoGraphOptimizer,
              timeout: Optional[timedelta] = None,
              num_iterations: Optional[int] = None):
    optimizer, objective = optimizer_setup(target_graph,
                                           optimizer_cls=optimizer_cls,
                                           timeout=timeout,
                                           num_iterations=num_iterations)
    found_graphs = optimizer.optimise(objective)
    found_graph = found_graphs[0] if isinstance(found_graphs, Sequence) else found_graphs
    history = optimizer.history
    return found_graph, history



from datetime import timedelta
from functools import partial
from typing import Type, Optional, Sequence

import networkx as nx

from examples.synthetic_graph_evolution.experiment import run_experiments
from examples.synthetic_graph_evolution.generators import generate_labeled_graph
from golem.core.adapter.nx_adapter import BaseNetworkxAdapter
from golem.core.dag.verification_rules import DEFAULT_DAG_RULES
from golem.core.optimisers.genetic.gp_optimizer import EvoGraphOptimizer
from golem.core.optimisers.genetic.gp_params import GPAlgorithmParameters
from golem.core.optimisers.genetic.operators.base_mutations import MutationTypesEnum
from golem.core.optimisers.genetic.operators.crossover import CrossoverTypesEnum
from golem.core.optimisers.objective import Objective
from golem.core.optimisers.optimization_parameters import GraphRequirements
from golem.core.optimisers.optimizer import GraphGenerationParams, GraphOptimizer
from golem.metrics.edit_distance import tree_edit_dist
from golem.metrics.graph_metrics import degree_distance


def tree_search_setup(target_graph: nx.DiGraph,
                      optimizer_cls: Type[GraphOptimizer] = EvoGraphOptimizer,
                      node_types: Sequence[str] = ('x',),
                      timeout: Optional[timedelta] = None,
                      num_iterations: Optional[int] = None):
    # Setup parameters
    num_nodes = target_graph.number_of_nodes()
    requirements = GraphRequirements(
        max_arity=num_nodes,
        max_depth=num_nodes,
        early_stopping_timeout=10,
        early_stopping_iterations=500,
        keep_n_best=4,
        max_graph_fit_time=timedelta(seconds=10),
        timeout=timeout,
        num_of_generations=num_iterations,
        n_jobs=-1,
        history_dir=None,
    )
    gp_params = GPAlgorithmParameters(
        multi_objective=False,
        mutation_types=[
            MutationTypesEnum.single_add,
            MutationTypesEnum.single_drop,
        ],
        crossover_types=[CrossoverTypesEnum.subtree]
    )
    graph_gen_params = GraphGenerationParams(
        adapter=BaseNetworkxAdapter(),
        rules_for_constraint=DEFAULT_DAG_RULES,
        available_node_types=node_types,
    )

    # Setup objective:
    # - primary metric is edit distance between 2 trees
    # - secondary metric is difference in node degree distribution
    objective = Objective(
        quality_metrics={'edit_dist': partial(tree_edit_dist, target_graph)},
        complexity_metrics={'degree': partial(degree_distance, target_graph)},
        is_multi_objective=gp_params.multi_objective
    )

    # Generate simple initial population with small tree graphs
    initial_graphs = [generate_labeled_graph('tree', 5, node_types)
                      for _ in range(gp_params.pop_size)]
    # Build the optimizer
    optimiser = optimizer_cls(objective, initial_graphs, requirements, graph_gen_params, gp_params)
    return optimiser, objective


if __name__ == '__main__':
    """
    In this experiment Optimizer is expected to find the target graph 
    (exact or almost exact, depending on available time and graph complexity)
    using Tree Edit Distance metric and a random tree (nx.random_tree) as target.
    
    This convergence can be seen from achieved metrics and visually from graph plots.
    """
    results_log = run_experiments(optimizer_setup=tree_search_setup,
                                  optimizer_cls=EvoGraphOptimizer,
                                  graph_names=['tree'],
                                  graph_sizes=[16],
                                  num_trials=1,
                                  trial_timeout=5,
                                  trial_iterations=100,
                                  visualize=True)
    print(results_log)



from typing import Dict, Callable, Collection, Sequence, Optional

import networkx as nx
import numpy as np

from examples.synthetic_graph_evolution.utils import draw_graphs_subplots

NumNodes = int
DiGraphGenerator = Callable[[NumNodes], nx.DiGraph]


graph_generators: Dict[str, DiGraphGenerator] = {
    'line': lambda n: nx.path_graph(n, create_using=nx.DiGraph),
    'tree': lambda n: nx.random_tree(n, create_using=nx.DiGraph),
    'gnp': lambda n: nx.gnp_random_graph(n, p=0.1),
    'star': nx.star_graph,
    '2ring': nx.circular_ladder_graph,
    'grid2d': lambda n: nx.grid_2d_graph(int(np.sqrt(n)), int(np.sqrt(n))),
    'hypercube': lambda n: nx.hypercube_graph(int(np.log2(n).round())),
}

graph_kinds: Sequence[str] = tuple(graph_generators.keys())


def nx_to_directed(graph: nx.Graph) -> nx.DiGraph:
    """Randomly chooses a direction for each edge."""
    if isinstance(graph, nx.DiGraph):
        return graph

    dedges = set()
    digraph = nx.DiGraph()

    for node, data in graph.nodes(data=True):
        digraph.add_node(node, **data)

    for u, v, data in graph.edges.data():
        edge = (u, v)
        inv_edge = (v, u)
        if edge in dedges or inv_edge in dedges:
            continue

        if np.random.default_rng().random() > 0.5:
            digraph.add_edge(*edge, **data)
            dedges.add(edge)
        else:
            digraph.add_edge(*inv_edge, **data)
            dedges.add(inv_edge)
    return digraph


def largest_component(graph: nx.Graph) -> nx.Graph:
    get_components = nx.weakly_connected_components \
        if isinstance(graph, nx.DiGraph) else nx.connected_components
    largest = max(get_components(graph), key=len)
    return graph.subgraph(largest)


def relabel_nx_graph(graph: nx.Graph, available_names: Collection[str]) -> nx.Graph:
    """Randomly label nodes with 'name' attribute in nx.Graph
    given list of available labels"""
    names = np.random.choice(available_names, size=graph.number_of_nodes())
    attributes = {node_id: {'name': name}
                  for node_id, name in zip(graph.nodes, names)}
    nx.set_node_attributes(graph, attributes)
    return graph


def postprocess_nx_graph(graph: nx.Graph,
                         node_labels: Optional[Sequence[str]] = ('x',),
                         connected: bool = True,
                         directed: bool = True):
    """Generate randomly labeled graph, enforce connectedness and direction."""
    # Remove unconnected components
    if connected:
        graph = largest_component(graph)
    # Optionally choose random directions for each edge
    if directed:
        # reverse() is mainly for trees
        # to make them growing towards root
        graph = nx_to_directed(graph).reverse()
    # Label the nodes randomly from available nodes
    if node_labels:
        graph = relabel_nx_graph(graph, node_labels)
    return graph


def generate_labeled_graph(kind: str,
                           size: int,
                           node_labels: Optional[Sequence[str]] = ('x',),
                           connected: bool = True,
                           directed: bool = True):
    """Generate randomly labeled graph of the specified kind and size,
    optionally enforce connectedness and direction."""
    graph = graph_generators[kind](size)
    return postprocess_nx_graph(graph, node_labels, connected, directed)


def _draw_sample_graphs(kind: str = 'gnp', sizes=tuple(range(5, 50, 5))):
    graphs = [generate_labeled_graph(kind, n) for n in sizes]
    draw_graphs_subplots(*graphs)


if __name__ == '__main__':
    _draw_sample_graphs('gnp')



from __future__ import annotations

import csv
import io
import itertools
import os
import shutil
from pathlib import Path
from typing import Any, Dict, List, Optional, Sequence, Union, TYPE_CHECKING

from golem.core.log import default_log
from golem.core.optimisers.objective.objective import ObjectiveInfo
from golem.core.optimisers.opt_history_objects.generation import Generation

from golem.core.paths import default_data_dir
from golem.serializers.serializer import default_load, default_save
from golem.visualisation.opt_viz import OptHistoryVisualizer

if TYPE_CHECKING:
    from golem.core.dag.graph import Graph
    from golem.core.optimisers.opt_history_objects.individual import Individual


class OptHistory:
    """
    Contains optimization history, save history to csv.
    Can be used for any type of graph that is serializable with Serializer.

    Args:
        objective: information about metrics (metric names and if it's multi-objective)
        default_save_dir: default directory used for saving history when not explicit path is provided.
    """

    def __init__(self,
                 objective: Optional[ObjectiveInfo] = None,
                 default_save_dir: Optional[os.PathLike] = None):
        self._objective = objective or ObjectiveInfo()
        self.individuals: List[Generation] = []
        self.archive_history: List[List[Individual]] = []
        self._tuning_result: Optional[Graph] = None
        self._log = default_log(self)

        # init default save directory
        if default_save_dir:
            default_save_dir = Path(default_save_dir)
            if not default_save_dir.is_absolute():
                # if path is not absolute, treat it as relative to data dir
                default_save_dir = Path(default_data_dir()) / default_save_dir
        else:
            default_save_dir = default_data_dir()
        self._default_save_dir = str(default_save_dir)

    @property
    def objective(self):
        return self._objective

    def is_empty(self) -> bool:
        return not self.individuals

    def add_to_history(self, individuals: Sequence[Individual], generation_label: Optional[str] = None,
                       generation_metadata: Optional[Dict[str, Any]] = None):
        generation = Generation(individuals, self.generations_count, generation_label, generation_metadata)
        self.individuals.append(generation)

    def add_to_archive_history(self, individuals: Sequence[Individual]):
        self.archive_history.append(list(individuals))

    def to_csv(self, save_dir: Optional[os.PathLike] = None, file: os.PathLike = 'history.csv'):
        save_dir = save_dir or self._default_save_dir
        if not os.path.isdir(save_dir):
            os.mkdir(save_dir)

        with open(Path(save_dir, file), 'w', newline='') as file:
            writer = csv.writer(file, quoting=csv.QUOTE_ALL)

            # Write header
            metric_str = 'metric'
            if self.objective.is_multi_objective:
                metric_str += 's'
            header_row = ['index', 'generation', metric_str, 'quantity_of_operations', 'depth', 'metadata']
            writer.writerow(header_row)

            # Write history rows
            idx = 0
            for gen_num, gen_inds in enumerate(self.individuals):
                for ind_num, ind in enumerate(gen_inds):
                    row = [idx, gen_num, ind.fitness.values, ind.graph.length, ind.graph.depth, ind.metadata]
                    writer.writerow(row)
                    idx += 1

    def save_current_results(self, save_dir: Optional[os.PathLike] = None):
        # Create folder if it's not exists
        save_dir = save_dir or self._default_save_dir
        if not os.path.isdir(save_dir):
            os.makedirs(save_dir)
            self._log.info(f"Created directory for saving optimization history: {save_dir}")

        try:
            last_gen_id = self.generations_count - 1
            last_gen = self.individuals[last_gen_id]
            for individual in last_gen:
                ind_path = Path(save_dir, str(last_gen_id), str(individual.uid))
                if not os.path.isdir(ind_path):
                    os.makedirs(ind_path)
                individual.save(json_file_path=Path(ind_path, f'{str(individual.uid)}.json'))
        except Exception as ex:
            self._log.exception(ex)

    def save(self, json_file_path: Union[str, os.PathLike] = None) -> Optional[str]:
        return default_save(obj=self, json_file_path=json_file_path)

    @staticmethod
    def load(json_str_or_file_path: Union[str, os.PathLike] = None) -> OptHistory:
        return default_load(json_str_or_file_path)

    @staticmethod
    def clean_results(dir_path: Optional[str] = None):
        """Clearn the directory tree with previously dumped history results."""
        if dir_path and os.path.isdir(dir_path):
            shutil.rmtree(dir_path, ignore_errors=True)
            os.mkdir(dir_path)

    @property
    def historical_fitness(self) -> Sequence[Sequence[Union[float, Sequence[float]]]]:
        """Return sequence of histories of generations per each metric"""
        if self.objective.is_multi_objective:
            historical_fitness = []
            num_metrics = len(self.individuals[0][0].fitness.values)
            for objective_num in range(num_metrics):
                # history of specific objective for each generation
                objective_history = [[ind.fitness.values[objective_num] for ind in generation]
                                     for generation in self.individuals]
                historical_fitness.append(objective_history)
        else:
            historical_fitness = [[ind.fitness.value for ind in pop]
                                  for pop in self.individuals]
        return historical_fitness

    @property
    def all_historical_fitness(self) -> List[float]:
        historical_fitness = self.historical_fitness
        if self.objective.is_multi_objective:
            all_historical_fitness = []
            for obj_num in range(len(historical_fitness)):
                all_historical_fitness.append(list(itertools.chain(*historical_fitness[obj_num])))
        else:
            all_historical_fitness = list(itertools.chain(*historical_fitness))
        return all_historical_fitness

    def all_historical_quality(self, metric_position: int = 0) -> List[float]:
        """
        Return fitness history of population for specified metric.

        Args:
            metric_position: Index of the metric for multi-objective optimization.
             By default, choose first metric, assuming it is primary quality metric.

        Returns:
            List: all historical fitness
        """
        if self.objective.is_multi_objective:
            all_historical_quality = self.all_historical_fitness[metric_position]
        else:
            all_historical_quality = self.all_historical_fitness
        return all_historical_quality

    @property
    def show(self):
        return OptHistoryVisualizer(self)

    def get_leaderboard(self, top_n: int = 10) -> str:
        """
        Prints ordered description of the best solutions in history
        :param top_n: number of solutions to print
        """
        # Take only the first graph's appearance in history
        individuals_with_positions \
            = list({ind.graph.descriptive_id: (ind, gen_num, ind_num)
                    for gen_num, gen in enumerate(self.individuals)
                    for ind_num, ind in reversed(list(enumerate(gen)))}.values())

        top_individuals = sorted(individuals_with_positions,
                                 key=lambda pos_ind: pos_ind[0].fitness, reverse=True)[:top_n]

        output = io.StringIO()
        separator = ' | '
        header = separator.join(['Position', 'Fitness', 'Generation', 'Graph'])
        print(header, file=output)
        for ind_num, ind_with_position in enumerate(top_individuals):
            individual, gen_num, ind_num = ind_with_position
            positional_id = f'g{gen_num}-i{ind_num}'
            print(separator.join([f'{ind_num:>3}, '
                                  f'{str(individual.fitness):>8}, '
                                  f'{positional_id:>8}, '
                                  f'{individual.graph.descriptive_id}']), file=output)

        # add info about initial assumptions (stored as zero generation)
        for i, individual in enumerate(self.individuals[0]):
            ind = f'I{i}'
            positional_id = '-'
            print(separator.join([f'{ind:>3}'
                                  f'{str(individual.fitness):>8}',
                                  f'{positional_id}',
                                  f'{individual.graph.descriptive_id}']), file=output)
        return output.getvalue()

    @property
    def initial_assumptions(self) -> Optional[Generation]:
        if not self.individuals:
            return None
        for gen in self.individuals:
            if gen.label == 'initial_assumptions':
                return gen

    @property
    def final_choices(self) -> Optional[Generation]:
        if not self.individuals:
            return None
        for gen in reversed(self.individuals):
            if gen.label == 'final_choices':
                return gen

    @property
    def generations_count(self) -> int:
        return len(self.individuals)

    @property
    def tuning_result(self):
        if hasattr(self, '_tuning_result'):
            return self._tuning_result
        else:
            return None

    @tuning_result.setter
    def tuning_result(self, val):
        self._tuning_result = val



from datetime import timedelta
from functools import partial
from typing import Type, Optional

from examples.synthetic_graph_evolution.experiment import run_experiments
from examples.synthetic_graph_evolution.generators import generate_labeled_graph
from golem.core.adapter.nx_adapter import BaseNetworkxAdapter
from golem.core.dag.verification_rules import has_no_self_cycled_nodes
from golem.core.optimisers.genetic.gp_optimizer import EvoGraphOptimizer
from golem.core.optimisers.genetic.gp_params import GPAlgorithmParameters
from golem.core.optimisers.genetic.operators.base_mutations import MutationTypesEnum
from golem.core.optimisers.genetic.operators.crossover import CrossoverTypesEnum
from golem.core.optimisers.genetic.operators.inheritance import GeneticSchemeTypesEnum
from golem.core.optimisers.objective import Objective
from golem.core.optimisers.optimization_parameters import GraphRequirements
from golem.core.optimisers.optimizer import GraphGenerationParams, GraphOptimizer
from golem.metrics.graph_metrics import *


def graph_search_setup(target_graph: nx.DiGraph,
                       optimizer_cls: Type[GraphOptimizer] = EvoGraphOptimizer,
                       node_types: Sequence[str] = ('X',),
                       timeout: Optional[timedelta] = None,
                       num_iterations: Optional[int] = None):
    # Setup parameters
    num_nodes = target_graph.number_of_nodes()
    requirements = GraphRequirements(
        max_arity=num_nodes,
        max_depth=num_nodes,
        early_stopping_timeout=5,
        early_stopping_iterations=1000,
        keep_n_best=4,
        timeout=timeout,
        num_of_generations=num_iterations,
        n_jobs=-1,
        history_dir=None,
    )
    gp_params = GPAlgorithmParameters(
        multi_objective=True,
        genetic_scheme_type=GeneticSchemeTypesEnum.parameter_free,
        mutation_types=[
            MutationTypesEnum.single_add,
            MutationTypesEnum.single_drop,
            MutationTypesEnum.single_edge,
        ],
    )
    graph_gen_params = GraphGenerationParams(
        adapter=BaseNetworkxAdapter(),
        rules_for_constraint=[has_no_self_cycled_nodes],
        available_node_types=node_types,
    )

    # Setup objective that measures some graph-theoretic similarity measure
    objective = Objective(
        quality_metrics={
            'sp_adj': partial(spectral_dist, target_graph, kind='adjacency'),
            'sp_lapl': partial(spectral_dist, target_graph, kind='laplacian'),
        },
        complexity_metrics={
            'degree': partial(degree_distance, target_graph),
            'graph_size': partial(size_diff, target_graph),
        },
        is_multi_objective=True
    )

    # Generate simple initial population with line graphs
    initial_graphs = [generate_labeled_graph('line', k+3, node_types)
                      for k in range(gp_params.pop_size)]
    # Build the optimizer
    optimiser = optimizer_cls(objective, initial_graphs, requirements, graph_gen_params, gp_params)
    return optimiser, objective


if __name__ == '__main__':
    results_log = run_experiments(optimizer_setup=graph_search_setup,
                                  optimizer_cls=EvoGraphOptimizer,
                                  graph_names=['2ring', 'gnp'],
                                  graph_sizes=[30, 100],
                                  num_trials=1,
                                  trial_timeout=5,
                                  trial_iterations=2000,
                                  visualize=True)
    print(results_log)



from copy import deepcopy
from random import choice, random
from typing import Callable, List, Union, Tuple, TYPE_CHECKING, Mapping, Hashable, Optional

import numpy as np

from golem.core.dag.graph import Graph
from golem.core.optimisers.genetic.operators.base_mutations import base_mutations_repo, MutationTypesEnum, MutationStrengthEnum
from golem.core.optimisers.genetic.operators.operator import PopulationT, Operator
from golem.core.optimisers.graph import OptGraph
from golem.core.optimisers.opt_history_objects.individual import Individual
from golem.core.optimisers.opt_history_objects.parent_operator import ParentOperator
from golem.core.optimisers.optimization_parameters import GraphRequirements
from golem.core.optimisers.optimizer import GraphGenerationParams, AlgorithmParameters

if TYPE_CHECKING:
    from golem.core.optimisers.genetic.gp_params import GPAlgorithmParameters


MutationFunc = Callable[[Graph, GraphRequirements, GraphGenerationParams, AlgorithmParameters], Graph]
MutationRepo = Mapping[Hashable, MutationFunc]


class Mutation(Operator):
    def __init__(self,
                 parameters: 'GPAlgorithmParameters',
                 requirements: GraphRequirements,
                 graph_gen_params: GraphGenerationParams,
                 mutations_repo: Optional[MutationRepo] = None
                 ):
        super().__init__(parameters, requirements)
        self.graph_generation_params = graph_gen_params
        self.parameters = parameters
        self._mutations_repo = mutations_repo or base_mutations_repo

    def __call__(self, population: Union[Individual, PopulationT]) -> Union[Individual, PopulationT]:
        if isinstance(population, Individual):
            return self._mutation(population)
        return list(map(self._mutation, population))

    def _mutation(self, individual: Individual) -> Individual:
        """ Function applies mutation operator to graph """

        for _ in range(self.parameters.max_num_of_operator_attempts):
            new_graph = deepcopy(individual.graph)
            num_mut = max(int(round(np.random.lognormal(0, sigma=0.5))), 1)

            new_graph, mutation_names = self._apply_mutations(new_graph, num_mut)

            is_correct_graph = self.graph_generation_params.verifier(new_graph)
            if is_correct_graph:
                parent_operator = ParentOperator(type_='mutation', operators=tuple(mutation_names),
                                                 parent_individuals=individual)
                return Individual(new_graph, parent_operator,
                                  metadata=self.requirements.static_individual_metadata)

        self.log.debug('Number of mutation attempts exceeded. '
                       'Please check optimization parameters for correctness.')

        return individual

    def _apply_mutations(self, new_graph: OptGraph, num_mut: int) -> Tuple[OptGraph, List[str]]:
        """Apply a number of mutations iteratively"""
        mutation_types = self.parameters.mutation_types
        is_static_mutation_type = random() < self.parameters.static_mutation_prob
        mutation_type = choice(mutation_types)
        mutation_names = []
        for _ in range(num_mut):
            # determine mutation type
            if not is_static_mutation_type:
                mutation_type = choice(mutation_types)
            is_custom_mutation = isinstance(mutation_type, Callable)

            new_graph, applied = self._adapt_and_apply_mutation(new_graph, mutation_type)

            if applied:
                mutation_names.append(str(mutation_type))  # log mutation
                if is_custom_mutation:  # custom mutation occurs once
                    break
        return new_graph, mutation_names

    def _adapt_and_apply_mutation(self, new_graph: OptGraph, mutation_type) -> Tuple[OptGraph, bool]:
        applied = self._will_mutation_be_applied(mutation_type)
        if applied:
            # get the mutation function and adapt it
            mutation_func = self._get_mutation_func(mutation_type)
            new_graph = mutation_func(new_graph, requirements=self.requirements,
                                      graph_gen_params=self.graph_generation_params,
                                      parameters=self.parameters)
        return new_graph, applied


    def _will_mutation_be_applied(self, mutation_type: Union[MutationTypesEnum, Callable]) -> bool:
        return random() <= self.parameters.mutation_prob and mutation_type is not MutationTypesEnum.none

    def _get_mutation_func(self, mutation_type: Union[MutationTypesEnum, Callable]) -> Callable:
        if isinstance(mutation_type, Callable):
            mutation_func = mutation_type
        else:
            mutation_func = self._mutations_repo[mutation_type]
        adapted_mutation_func = self.graph_generation_params.adapter.adapt_func(mutation_func)
        return adapted_mutation_func



from random import shuffle
from typing import TYPE_CHECKING

from golem.core.optimisers.genetic.operators.operator import PopulationT, Operator
from golem.core.utilities.data_structures import ComparableEnum as Enum


class ElitismTypesEnum(Enum):
    keep_n_best = 'keep_n_best'
    replace_worst = 'replace_worst'
    none = 'none'


class Elitism(Operator):
    def __call__(self, best_individuals: PopulationT, new_population: PopulationT) -> PopulationT:
        elitism_type = self.parameters.elitism_type
        if elitism_type is ElitismTypesEnum.none or not self._is_elitism_applicable():
            return new_population
        elif elitism_type is ElitismTypesEnum.keep_n_best:
            return self._keep_n_best_elitism(best_individuals, new_population)
        elif elitism_type is ElitismTypesEnum.replace_worst:
            return self._replace_worst_elitism(best_individuals, new_population)
        else:
            raise ValueError(f'Required elitism type not found: {elitism_type}')

    def _is_elitism_applicable(self) -> bool:
        if self.parameters.multi_objective:
            return False
        return self.parameters.pop_size >= self.parameters.min_pop_size_with_elitism

    def _keep_n_best_elitism(self, best_individuals: PopulationT, new_population: PopulationT) -> PopulationT:
        shuffle(new_population)
        new_population[:len(best_individuals)] = best_individuals
        return new_population

    def _replace_worst_elitism(self, best_individuals: PopulationT, new_population: PopulationT) -> PopulationT:
        population = best_individuals + new_population
        # sort in descending order (Fitness(10) > Fitness(11))
        sorted_ascending_population = sorted(population, key=lambda individual: individual.fitness, reverse=True)
        return sorted_ascending_population[:len(new_population)]
